# coding: utf8
from __future__ import unicode_literals

import numpy

from ..api import layerize
from .ops import NumpyOps, CupyOps
from ._classes.model import Model


try:
    from cupy import get_array_module
except ImportError:
    get_array_module = lambda arr: numpy


def Pooling(*funcs, **kwargs):
    ops = kwargs["ops"] if "ops" in kwargs else Model.ops
    F = len(funcs)
    drop_factor = kwargs.get("drop_factor", 1.0)

    def begin_update(X_lengths, drop=0.0):
        if drop is not None:
            drop *= drop_factor
        X, lengths = X_lengths
        T, O = X.shape  # noqa: E741
        pooled = ops.allocate((len(lengths), F * O))
        bp_funcs = [None] * F
        for i, func in enumerate(funcs):
            res, bp_res = func.begin_update((X, lengths))
            pooled[:, i * O : i * O + O] = res
            bp_funcs[i] = bp_res
        pooled, bp_dropout = ops.dropout(pooled, drop)

        def finish_update(d_pooled, sgd=None):
            dX = ops.allocate(X.shape)
            for i, bp_func in enumerate(bp_funcs):
                dX += bp_func(d_pooled[:, i * O : i * O + O])
            return dX

        return pooled, bp_dropout(finish_update)

    return layerize(begin_update)


@layerize
def mean_pool(X_lengths, drop=0.0):
    X, lengths = X_lengths
    if isinstance(X, numpy.ndarray):
        ops = NumpyOps()
    else:
        ops = CupyOps()
    output = ops.mean_pool(X, lengths)

    def finish_update(d_output, sgd=None):
        d_output = ops.xp.ascontiguousarray(d_output)
        return ops.backprop_mean_pool(d_output, lengths)

    return output, finish_update


@layerize
def sum_pool(X_lengths, drop=0.0):
    X, lengths = X_lengths
    if isinstance(X, numpy.ndarray):
        ops = NumpyOps()
    else:
        ops = CupyOps()

    output = ops.sum_pool(X, lengths)

    def finish_update(d_output, sgd=None):
        d_output = ops.xp.ascontiguousarray(d_output)
        return ops.backprop_sum_pool(d_output, lengths)

    return output, finish_update


@layerize
def max_pool(X_lengths, drop=0.0):
    X, lengths = X_lengths
    ops = Model.ops

    best, which = ops.max_pool(X, lengths)

    def finish_update(d_output, sgd=None):
        d_output = ops.xp.ascontiguousarray(d_output)
        return ops.backprop_max_pool(d_output, which, lengths)

    return best, finish_update



# coding: utf8
from __future__ import unicode_literals

from .affine import Affine
from ... import check
from ...check import has_shape


class ReLu(Affine):
    @check.arg(1, has_shape(("nB", "nI")))
    def predict(self, input__BI):
        output__BO = Affine.predict(self, input__BI)
        output__BO = self.ops.relu(output__BO, inplace=False)
        return output__BO

    @check.arg(1, has_shape(("nB", "nI")))
    def begin_update(self, input__BI, drop=0.0):
        output__BO, finish_affine = Affine.begin_update(self, input__BI, drop=0.0)
        output__BO = self.ops.relu(output__BO)

        @check.arg(0, has_shape(("nB", "nO")))
        def finish_update(gradient, sgd=None):
            gradient = self.ops.backprop_relu(gradient, output__BO)
            return finish_affine(gradient, sgd)

        dropped, bp_dropout = self.ops.dropout(output__BO, drop, inplace=False)
        return dropped, bp_dropout(finish_update)



# coding: utf8
from __future__ import unicode_literals

from .model import Model


class FeatureExtracter(Model):
    def __init__(self, attrs):
        Model.__init__(self)
        self.attrs = attrs

    def begin_update(self, docs, drop=0.0):
        # Handle spans
        features = [self._get_feats(doc) for doc in docs]
        return features, _feature_extracter_bwd

    def _get_feats(self, doc):
        if hasattr(doc, "to_array"):
            arr = doc.to_array(self.attrs)
        else:
            arr = doc.doc.to_array(self.attrs)[doc.start : doc.end]
        return self.ops.asarray(arr, dtype="uint64")


def _feature_extracter_bwd(d_features, sgd=None):
    return d_features



# coding: utf8
from __future__ import unicode_literals

import pytest
import numpy

from thinc.neural._classes.affine import Affine
from thinc.exceptions import ShapeMismatchError


def test_mismatched_shapes_raise_ShapeMismatchError():
    X = numpy.ones((3, 4))
    model = Affine(10, 5)
    with pytest.raises(ShapeMismatchError):
        model.predict(X)



# coding: utf8
from __future__ import unicode_literals

from .model import Model


class Residual(Model):
    def __init__(self, layer):
        Model.__init__(self)
        self._layers.append(layer)
        self.on_data_hooks.append(on_data)

    @property
    def nO(self):
        return self._layers[-1].nO

    def predict(self, X):
        Y = self._layers[0](X)
        if isinstance(X, list) or isinstance(X, tuple):
            return [X[i] + Y[i] for i in range(len(X))]
        elif isinstance(X, tuple) and isinstance(Y, tuple) and len(X) == 2:
            assert X[1].sum() == Y[1].sum()
            assert Y[1].sum() == Y[0].shape[0], (Y[1].sum(), Y[0].shape[0])
            return (X[0] + Y[0], Y[1])
        else:
            return X + Y

    def begin_update(self, X, drop=0.0):
        y, bp_y = self._layers[0].begin_update(X, drop=drop)
        if isinstance(X, list):
            output = [X[i] + y[i] for i in range(len(X))]
        elif isinstance(X, tuple) and isinstance(y, tuple) and len(X) == 2:
            # Handle case where we have (data, lengths) tuple
            assert X[1].sum() == y[1].sum()
            assert y[1].sum() == y[0].shape[0], (y[1].sum(), y[0].shape[0])
            output = (X[0] + y[0], y[1])
        else:
            output = X + y

        def residual_bwd(d_output, sgd=None):
            dX = bp_y(d_output, sgd)
            if isinstance(d_output, list) or isinstance(d_output, tuple):
                return [d_output[i] + dX[i] for i in range(len(d_output))]
            else:
                return d_output + dX

        return output, residual_bwd


def on_data(self, X, y=None):
    for layer in self._layers:
        for hook in layer.on_data_hooks:
            hook(layer, X, y)
        if hasattr(layer, "W"):
            layer.W.fill(0)



# coding: utf8
from __future__ import unicode_literals

from thinc.i2v import *  # noqa: F401, F403
from thinc.v2v import *  # noqa: F401, F403
from thinc.t2t import *  # noqa: F401, F403
from thinc.t2v import *  # noqa: F401, F403
from thinc.api import *  # noqa: F401, F403
from thinc.misc import *  # noqa: F401, F403



# coding: utf8
from __future__ import unicode_literals

from ._classes.model import Model  # noqa: F401
from ._classes.affine import Affine  # noqa: F401
from ._classes.relu import ReLu  # noqa: F401
from ._classes.elu import ELU  # noqa: F401
from ._classes.maxout import Maxout  # noqa: F401
from ._classes.softmax import Softmax  # noqa: F401
from ._classes.resnet import Residual  # noqa: F401



from __future__ import unicode_literals

from thinc.config import Config
from thinc.neural.optimizers import Optimizer


EXAMPLE_CONFIG = """
[DEFAULT]

[optimizer]
@optimizers = "Adam.v1"
beta1 = 0.9
beta2 = 0.999
use_averages = true

[optimizer.learn_rate]
@schedules = "warmup_linear.v1"
start = 0.1
steps = 10000

[pipeline]

[pipeline.parser]
name = "parser"
factory = "parser"

[pipeline.parser.model]
@layers = "spacy.ParserModel.v1"
hidden_depth = 1
hidden_width = 64
token_vector_width = 128

[pipeline.parser.model.tok2vec]
@layers = "Tok2Vec.v1"
width = ${pipeline.parser.model:token_vector_width}

[pipeline.parser.model.tok2vec.embed]
@layers = "spacy.MultiFeatureHashEmbed.v1"
width = ${pipeline.parser.model.tok2vec:width}

[pipeline.parser.model.tok2vec.embed.hidden]
@layers = "MLP.v1"
depth = 1
pieces = 3
layer_norm = true
outputs = ${pipeline.parser.model.tok2vec.embed:width}

[pipeline.parser.model.tok2vec.encode]
@layers = "spacy.MaxoutWindowEncoder.v1"
depth = 4
pieces = 3
window_size = 1

[pipeline.parser.model.lower]
@layers = "spacy.ParserLower.v1"

[pipeline.parser.model.upper]
@layers = "thinc.Affine.v1"
"""

OPTIMIZER_CFG = """
[optimizer]
@optimizers = "Adam.v1"
beta1 = 0.9
beta2 = 0.999
use_averages = true

[optimizer.schedules.learn_rate]
@schedules = "warmup_linear.v1"
initial_rate = 0.1
warmup_steps = 10000
total_steps = 100000
"""

def test_read_config():
    byte_string = EXAMPLE_CONFIG.encode("utf8")
    cfg = Config().from_bytes(byte_string)
    assert cfg["optimizer"]["learn_rate"]["start"] == 0.1
    assert cfg["pipeline"]["parser"]["factory"] == "parser"
    assert cfg["pipeline"]["parser"]["model"]["tok2vec"]["width"] == 128


def test_optimizer_config():
    cfg = Config().from_bytes(OPTIMIZER_CFG.encode("utf8"))
    optimizer = Optimizer.from_config(cfg)

